package w2phtml.xhtml.content;

import static w2phtml.office.XMLString.TEXT_CONTINUE_NUMBERING;
import static w2phtml.office.XMLString.TEXT_H;
import static w2phtml.office.XMLString.TEXT_LIST;
import static w2phtml.office.XMLString.TEXT_LIST_HEADER;
import static w2phtml.office.XMLString.TEXT_LIST_ITEM;
import static w2phtml.office.XMLString.TEXT_ORDERED_LIST;
import static w2phtml.office.XMLString.TEXT_P;
import static w2phtml.office.XMLString.TEXT_RESTART_NUMBERING;
import static w2phtml.office.XMLString.TEXT_START_VALUE;
import static w2phtml.office.XMLString.TEXT_STYLE_NAME;
import static w2phtml.office.XMLString.TEXT_UNORDERED_LIST;

import java.util.Hashtable;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import w2phtml.office.ListCounter;
import w2phtml.office.ListStyle;
import w2phtml.office.OfficeReader;
import w2phtml.util.Misc;
import w2phtml.xhtml.Converter;
import w2phtml.xhtml.Parser;
import w2phtml.xhtml.StyleInfo;
import w2phtml.xhtml.XhtmlConfig;

public class ListParser extends Parser {

	private Hashtable<String, ListCounter> listCounters = new Hashtable<String, ListCounter>();
	protected String sCurrentListLabel = null;
	protected ListStyle currentListStyle = null;
	protected int nCurrentListLevel = 0;
  ListCounter outlineNumbering;


	public ListParser(OfficeReader ofr, XhtmlConfig config, Converter converter) {
		super(ofr, config, converter);
    outlineNumbering = new ListCounter(ofr.getOutlineStyle());
	}

	protected ListCounter getListCounter(ListStyle style) {
	    if (style==ofr.getOutlineStyle()) {
	        // Outline numbering has a special counter
	        return outlineNumbering;
	    }
	    else if (style!=null) {
	        // Get existing or create new counter
	        if (listCounters.containsKey(style.getName())) {
	            return listCounters.get(style.getName());
	        }
	        else {
	            ListCounter counter = new ListCounter(style);
	            listCounters.put(style.getName(),counter);
	            return counter;
	        }
	    }
	    else {
	        // No style, return a dummy
	        return new ListCounter();
	    }
	}

	protected void insertListLabel(ListStyle style, int nLevel, String sDefaultStyle, String sPrefix, String sLabel, Element hnode) {
	    if (sLabel!=null && sLabel.length()>0) {
	    	if (sPrefix!=null) {
	    		Element prefix = converter.createElement("span");
	    		prefix.setAttribute("class", "chapter-name");
	    		hnode.appendChild(prefix);
	    		prefix.appendChild( converter.createTextNode(sPrefix));
	    	}
	        StyleInfo info = new StyleInfo();
	        if (style!=null) {
	            String sTextStyleName = style.getLevelProperty(nLevel,TEXT_STYLE_NAME);
	            getTextSP().readStyle(sTextStyleName, info);
	        }
	
	        if (info.sTagName==null) { info.sTagName = "span"; }
	        if (info.sClass==null) { info.sClass = sDefaultStyle; }
	
	        Element content = converter.createElement(info.sTagName);
	        getTextSP().writeStyle(info, content);
	        hnode.appendChild(content);
	        content.appendChild( converter.createTextNode(sLabel) );
	    }
	}

	private boolean hasItems(Node onode) {
	    Node child = onode.getFirstChild();
	    while (child!=null) {
	        if (Misc.isElement(child,TEXT_LIST_ITEM) ||
	            Misc.isElement(child,TEXT_LIST_HEADER)) {
	            return true;
	        }
	        child = child.getNextSibling();
	    }
	    return false;
	}

	private void handleOL(Node onode, int nLevel, String sStyleName, Node hnode) {
	    if (hasItems(onode)) {
	        // add an OL element
	        Element list = converter.createElement("ol");
	        StyleInfo listInfo = new StyleInfo();
	        getListSP().applyStyle(nLevel,sStyleName,listInfo);
	        getListSP().readParentStyle(nLevel, sStyleName, listInfo);
	        writeStyle(listInfo,list);
	        hnode.appendChild(list);
	        traverseList(onode,nLevel,sStyleName,list);
	    }
	}

	private void handleUL(Node onode, int nLevel, String sStyleName, Node hnode) {
	    if (hasItems(onode)) {
	        // add an UL element
	        Element list = converter.createElement("ul");
	        StyleInfo listInfo = new StyleInfo();
	        getListSP().applyStyle(nLevel,sStyleName,listInfo);
	        writeStyle(listInfo,list);
	        hnode.appendChild(list);
	        traverseList(onode,nLevel,sStyleName,list);
	    }
	}

	protected void handleList(Node onode, int nLevel, String sStyleName, Node hnode) {
	    // In OpenDocument, we should use the style to determine the type of list
	    String sStyleName1 = Misc.getAttribute(onode,TEXT_STYLE_NAME);
	    if (sStyleName1!=null) { sStyleName = sStyleName1; }
	    ListStyle style = ofr.getListStyle(sStyleName);
	    if (style!=null && style.isNumber(nLevel)) {
	        handleOL(onode,nLevel,sStyleName,hnode);
	    }
	    else {
	        handleUL(onode,nLevel,sStyleName,hnode);
	    }
	}

	private void traverseList(Node onode, int nLevel, String styleName, Element hnode) {
	    ListCounter counter = getListCounter(ofr.getListStyle(styleName));
	
	    // Restart numbering, if required
	    //if (counter!=null) {
	    boolean bContinueNumbering = "true".equals(Misc.getAttribute(onode,TEXT_CONTINUE_NUMBERING));
	    if (!bContinueNumbering && counter!=null) {
	        counter.restart(nLevel);
	    }
	    if (config.listFormatting()==XhtmlConfig.CSS1_HACK && counter.getValue(nLevel)>0) {
	        hnode.setAttribute("start",Integer.toString(counter.getValue(nLevel)+1));                	
	    }
	    //}
	
	    if (onode.hasChildNodes()) {
	        NodeList nList = onode.getChildNodes();
	        int len = nList.getLength();
	        
	        for (int i = 0; i < len; i++) {
	            Node child = nList.item(i);
	            
	            if (child.getNodeType() == Node.ELEMENT_NODE) {
	                String nodeName = child.getNodeName();
	                
	                if (nodeName.equals(TEXT_LIST_ITEM)) {
	                    // Check to see if first child is a new list
	                    boolean bIsImmediateNestedList = false;
	                    Element child1 = Misc.getFirstChildElement(child);
	                    if (child1.getTagName().equals(TEXT_ORDERED_LIST) || // old
	                        child1.getTagName().equals(TEXT_UNORDERED_LIST) || // old
	                        child1.getTagName().equals(TEXT_LIST)) { // oasis
	                        bIsImmediateNestedList = true;
	                    }
	
	                    if (config.listFormatting()==XhtmlConfig.CSS1_HACK && bIsImmediateNestedList) {
	                        traverseListItem(child,nLevel,styleName,hnode);
	                    }
	                    else {
	                        // add an li element
	                    	//if (counter!=null) {
	                    	sCurrentListLabel = counter.step(nLevel).getLabel();
	                    	//}
	                        currentListStyle = ofr.getListStyle(styleName);
	                        nCurrentListLevel = nLevel;
	                        Element item = converter.createElement("li");
	                        StyleInfo info = new StyleInfo();
	                        getPresentationSP().applyOutlineStyle(nLevel,info);
	                        writeStyle(info,item);
	                        hnode.appendChild(item);
	                        if (config.listFormatting()==XhtmlConfig.CSS1_HACK) {
	                            boolean bRestart = "true".equals(Misc.getAttribute(child,
	                                TEXT_RESTART_NUMBERING));
	                            int nStartValue = Misc.getPosInteger(Misc.getAttribute(child,
	                                TEXT_START_VALUE),1);
	                            if (bRestart) {
	                                item.setAttribute("value",Integer.toString(nStartValue));
	                                //if (counter!=null) {
	                                sCurrentListLabel = counter.restart(nLevel,nStartValue).getLabel();
	                                //}
	                            }
	                        }
	                        traverseListItem(child,nLevel,styleName,item);
	                    }
	                }
	                if (nodeName.equals(TEXT_LIST_HEADER)) {
	                    // add an li element
	                    Element item = converter.createElement("li");
	                    hnode.appendChild(item);
	                    item.setAttribute("style","list-style-type:none");
	                    traverseListItem(child,nLevel,styleName,item);
	                }
	            }
	        }
	    }
	}

	private void traverseListItem(Node onode, int nLevel, String styleName, Node hnode) { 
	    
	    // Still here? - traverse block text as usual!
	    //TODO:
	    getTextParser().parseText(onode,nLevel,styleName,hnode);
	}

	protected boolean listIsOnlyHeadings(Node node) {
	    Node child = node.getFirstChild();
	    while (child!=null) {
	        if (child.getNodeType() == Node.ELEMENT_NODE) {
	            String nodeName = child.getNodeName();
	            if (nodeName.equals(TEXT_LIST_ITEM)) {
	                if (!itemIsOnlyHeadings(child)) return false;
	            }
	            else if (nodeName.equals(TEXT_LIST_HEADER)) {
	                if (!itemIsOnlyHeadings(child)) return false;
	            }
	        }
	        child = child.getNextSibling();
	    }
	    return true;
	}

	private boolean itemIsOnlyHeadings(Node node) {
	    Node child = node.getFirstChild();
	    while (child!=null) {
	        if (child.getNodeType() == Node.ELEMENT_NODE) {
	            String nodeName = child.getNodeName();
	            if (nodeName.equals(TEXT_LIST)) {
	                if (!listIsOnlyHeadings(child)) return false;
	            }
	            else if (nodeName.equals(TEXT_ORDERED_LIST)) {
	                if (!listIsOnlyHeadings(child)) return false;
	            }
	            else if (nodeName.equals(TEXT_UNORDERED_LIST)) {
	                if (!listIsOnlyHeadings(child)) return false;
	            }
	            else if(!nodeName.equals(TEXT_H)) {
	                return false;
	            }
	        }
	        child = child.getNextSibling();
	    }
	    return true;
	}

	protected Node handleFakeList(Node onode, int nLevel, String sStyleName, Node hnode) {
	    String sStyleName1 = Misc.getAttribute(onode,TEXT_STYLE_NAME);
	    if (sStyleName1!=null) { sStyleName = sStyleName1; }
	    return traverseFakeList(onode,hnode,nLevel,sStyleName);
	}

	private Node traverseFakeList(Node onode, Node hnode, int nLevel, String sStyleName) {
	    // Restart numbering?
	    boolean bContinueNumbering ="true".equals(
	        Misc.getAttribute(onode,TEXT_CONTINUE_NUMBERING));
	    if (!bContinueNumbering) {
	        getListCounter(ofr.getListStyle(sStyleName)).restart(nLevel);
	    }
	
	    Node child = onode.getFirstChild();
	    while (child!=null) {
	        if (child.getNodeType() == Node.ELEMENT_NODE) {
	            String sNodeName = child.getNodeName();
	            
	            if (sNodeName.equals(TEXT_LIST_ITEM)) {
	                boolean bRestart = "true".equals(Misc.getAttribute(child,
	                    TEXT_RESTART_NUMBERING));
	                int nStartValue = Misc.getPosInteger(Misc.getAttribute(child,
	                    TEXT_START_VALUE),1);
	                hnode = traverseFakeListItem(child, hnode, nLevel, sStyleName, false, bRestart, nStartValue);
	            }
	            else if (sNodeName.equals(TEXT_LIST_HEADER)) {
	                hnode = traverseFakeListItem(child, hnode, nLevel, sStyleName, true, false, 0);
	            }
	        }
	        child = child.getNextSibling();
	    }
	    return hnode;
	}

	private Node traverseFakeListItem(Node onode, Node hnode, int nLevel, String sStyleName, boolean bUnNumbered, boolean bRestart, int nStartValue) {
		Node child = onode.getFirstChild();
		while (child != null) {
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				String sNodeName = child.getNodeName();
	
				if (sNodeName.equals(TEXT_H)) {
					Node rememberNode = hnode;
					getTextParser().handleHeading((Element) child, (Element) hnode, rememberNode != hnode, ofr.getListStyle(sStyleName), nLevel, bUnNumbered, bRestart, nStartValue);
				} else if (sNodeName.equals(TEXT_P)) {
					// Currently we only handle fakes lists containing headings
				} else if (sNodeName.equals(TEXT_LIST)) { // oasis
					return traverseFakeList(child, hnode, nLevel + 1, sStyleName);
				} else if (sNodeName.equals(TEXT_ORDERED_LIST)) { // old
					return traverseFakeList(child, hnode, nLevel + 1, sStyleName);
				} else if (sNodeName.equals(TEXT_UNORDERED_LIST)) { // old
					return traverseFakeList(child, hnode, nLevel + 1, sStyleName);
				}
			}
			child = child.getNextSibling();
		}
		return hnode;
	}

}